/* LSQ.C
	Routine to fit spectral data to a set of
	Gaussians by chi-squared minimization,
	using quasi-linear regression

	For MCA program
	(c) Copyright  D.W.Curtis  1987,91,95
	Converted to MSC  10-95  DWC


	Y[nY] are the data points;
	Sig[nY] are the mean squared deviations for Y

	P1[nP] = parameters of fit

	P0[nF] = first parameter associated with function nF
	(*func[nF])(x,P1) = array of pointers to fit routines
				(P1 = sub-array of parameters for func)
	(*dfunc[nF])(x,iP,P1) = array of pointers to routines to
				compute partial derivative of func wrt parameter iP
				at point x

	Fit((float)i,P1) = fit function at point i with parameters P1
		= Sum(nF) (*func[iF])((float)i,&P1[P0[iF]])

	dFdP(n,i) = partial derivative of fit function wrt parameter # n
				at point i (with current parameters P)

	returns the chi-squared of the fit


	Note that initial guesses of parameters are required
*/

#include <stdio.h>

#define LAMBDA0 1.0F		/* Non-linearity parameters (gradient vs regression mix) */
#define LAMBDAS 3.0F

extern float Y[], Sig[], P1[];
extern int nY,nP;

extern int nF, P0[];
extern float (*func[])(float x, float P[]), (*dfunc[])(int iX, int iP, float P[]);

#define SIZE 17
#define NITT 300

static void regress(float RM[], float RB[]);
static float chisq(float A[]);
static float Fit(float x, float P1[]);
static float dFdP(int ip, int ix);

float lsq()		/* do minimization, return chi-squared */
{
	float RM[SIZE*SIZE],RMI[SIZE*SIZE],RB[SIZE],A[SIZE];
	float det,chisq0,chisq1,dcs;
	void err_out(char *Str),mvmult(float *R, float V[], float *A, int N);
	float smatinv(float R[], int N);
	float Lambda;
	int i,j,l,first,itt;
	static char str[50];

	for(i=0; i<nP; i++) A[i] = P1[i];

	chisq0 = chisq(A);

	Lambda = LAMBDA0;
	dcs = 1.F;
	itt = 0;

	while(((dcs > .01) || (Lambda > 0.001)) && itt < NITT) {	/* Update fit */
		regress(RM,RB);
		first = 1;
		while(first || dcs < -.05F) {	/* increase Lambda till chisq gets smaller */
			if(first) first=0;
			else if((Lambda *= LAMBDAS) > 50.F) {
				err_out("Not converging");
				return(chisq0);
			}
			sprintf(str,"Itt %d, X^2=%8.1f",itt,chisq0);
			err_out(str);
			for(l=i=0; i< nP; i++) for(j=0; j<nP; j++,l++) {
				if(i==j) RMI[l] = RM[l]*(1.F+Lambda);	/* Modify gradient */
				else RMI[l] = RM[l];
			}
			det = smatinv(RMI,nP);
			if(det == 0.F) {
				err_out("LSQ error");
				dcs = -1.F;
			}
			else {
				mvmult(RMI,RB,A,nP);
				for(i=0; i<nP; i++) A[i] += P1[i];		/* Update estimate */
				dcs = (chisq0-(chisq1=chisq(A)))/chisq0;
			}
		}
		Lambda /= LAMBDAS;		/* Getting better - use less gradient */

		for(i=0; i<nP; i++) P1[i] = A[i];
		chisq0 = chisq1;

		if(dcs < 0.F) dcs = -dcs;
//		if(ci_ready()) {
//			err_out("*Break*");
//			break;
//		}
		itt++;
	}

/*	det = smatinv(RM,nP);		/* Lambda = 0. 
	if(det != 0.) {
		mvmult(RM,RB,A,nP);
		for(i=0; i<nP; i++) P1[i] += A[i];
		chisq0 = chisq(P);
	}
*/
//	if(ci_ready()) ci_scan();
//	else err_out("");
	return(chisq0);
}


/* Routine to compute regression matrix RM and vector RB
		RMnm = Sum { (dFdPn * dFdPm) / Y }
		RBn = Sum { (Y-F) * dFdPn / Y }

*/

static void regress(float RM[], float RB[])
{
	int iY,iA1,iA2;
	float *RMptr,dfdp[SIZE],d;

	for(iA1=0; iA1<nP; iA1++) {		/* clear regression matrix */
		RB[iA1] = 0.F;
		RMptr = &RM[iA1*nP];
		for(iA2=0; iA2<=iA1; iA2++) *RMptr++ = 0.F;
	}
	
	for(iY=0; iY<nY; iY++) if(Sig[iY] < 1.e10F) {
		d = (Y[iY] - Fit((float)iY,P1))/Sig[iY];
		for(iA1=0; iA1 < nP; iA1++) {
			RB[iA1] += d * (dfdp[iA1]=dFdP(iA1,iY));
			RMptr = &RM[iA1*nP];
			for(iA2=0; iA2<=iA1; iA2++)
				*RMptr++ += dfdp[iA1]*dfdp[iA2]/Sig[iY];
		}
	}

	for(iA1=0; iA1<nP; iA1++) {
		if(RM[iA1*nP + iA1] == 0.) RM[iA1*nP + iA1] = 1.e-4;
		for(iA2=0; iA2<iA1; iA2++)
			RM[iA2*nP + iA1] = RM[iA1*nP + iA2];
	}
}

		

/*
	Routine to compute Chi-squared Function:

	X = Sum ( (Yi - YFITi)**2 / SIGi)

*/

static float chisq(float A[])
{
	int iY;
	float ch,d,Fit(float Y, float A[]);

	ch = 0.F;
	for(iY=0; iY<nY; iY++) if(Sig[iY] < 1.e20F) {
		d = Y[iY] - Fit((float)iY,A);
		ch += d*d/Sig[iY];
	}
	return(ch);
}

static float Fit(float x, float P1[])
{
	int iF;
	float fit;

	fit = 0.F;
	for(iF=0; iF<nF; iF++) {
		fit += (*func[iF])(x,&P1[P0[iF]]);
	}
	return(fit);
}

static float dFdP(int ip, int ix)
{
	int iF;
	float df;

	for(iF=0; iF<nF && ip >= P0[iF]; iF++);
	iF--;
	df =(*dfunc[iF])(ix,ip-P0[iF],&P1[P0[iF]]);
	return(df);
}

	
